/**
 * @brief
 *    Demo for the Timer 1 in CTC mode.
 * @author
 *    Angel Grover Pérez Muñoz, Manuel Nieto
 * 
 * -- Timer1 properties --------------------------------------------------------
 *
 * - Clock runs at 16 MHz
 * - Prescaler configured by bits CS1[2:0] at register TCCR1B
 *       /8=010, /64=011, /256=100, /1024=101
 * - Counter Register: TCNT1
 * - Compare Register: OCR1A, OCR1B, OCR1C
 * - TOP Value: OCR1A, ICR1A, or fixed values (0x00FF, 0x01FF, 0x03FF, 0xFFFF).
 *       The TOP the value is also known as the resolution of the timer.
 *       If OCR1A is used, it can't be used to generate a PWM output.
 * - IRQ mask Register: TIMSK1
 *       TIMSK1{OCIE1A, OCIE1B, OCIE1C} to enable IRQ when TCNT1 == OCR1[A,B,C]
 *       TIMSK1{TOIE1} to enable IRQ at overflow (TCNT1 == TOP)
 * - Match flags:  TIFR1{OCF1A, OCF1B, OCF1C}
 * - IRQ flag Register: TIFR1
 *       Overflow flag: TIFR1{TOV1}
 *       Capture flags: TIFR1{ICF1}
 * - Out pin for Channel A: D9  (OC1A)
 * - Out pin for Channel B: D10 (OC1B)
 * - Out pin for Channel C: D11 (OC1C)
 *
 * - Counting sequence defined by TCCR1A{WGM1[1:0]} & TCCR1B{WGM1[2:3]}
 * - COMnx1:0 bits for defining the Output Compare (OCnx) state at the next compare match
 *
 * Normal mode:
 * - WGM1[2:0] = 0
 * - COM1x[1:0] = X for set, Y for clear, Z for toggle at compare match.
 *
 * CTC mode:
 * - COM1x[1:0] = 1 for toggle, 2 for clear, 3 for set at compare match.
 * - See table 14-5
 */

#include "./bitwise.h" // IMPORT my bitwise operations

/***************************************************************************//**
 * L E D   F U N C T I O N S
 ******************************************************************************/

#define LED_DDRX   (DDRC)
#define LED_DDRXn  (DDC7)
#define LED_PORTX  (PORTC)
#define LED_PORTXn (PORTC7)

/**
 * Configures the LED pin as an output I/O port.
 */
static void led_setup() {
  sbi(LED_DDRX, LED_DDRXn);
}

/**
 * A toggle turns the LED on if it is off, and turns it off if it is on.
 */
static void led_toggle() {
  static bool on = false;
  if (on) {
    cbi(LED_PORTX, LED_PORTXn); //< turn OFF (sets to 1, 5V)
    on = false;
  } else {
    sbi(LED_PORTX, LED_PORTXn); //< turn ON  (sets to 0, 0V)
    on = true;
  }
}

/***************************************************************************//**
 * T I M E R 1   F U N C T I O N S
 ******************************************************************************/

/**
 * Timer1 pins:
 */
#define D9_DDRX   (DDRB)
#define D9_DDRXn  (DDB5)
#define D10_DDRX  (DDRB)
#define D10_DDRXn (DDB6)
#define D11_DDRX  (DDRB)
#define D11_DDRXn (DDB7)

/**
 * Software counters for the Timer1.
 * These variables are updated inside TIMER1_COMPA_vect, TIMER1_COMPB_vect ISR
 * and read outside in the loop() function.
 * Therefore, to avoid corruptions the reading and writing operations outside ISRs must be atomic.
 */
volatile int64_t t1_count_a = 0;
volatile int64_t t1_count_b = 0;

/**
 * This is an atomic READ operation and it is intended to be used outside ISRs,
 * for example, in the loop() function.
 */
int64_t read_t1_count_a() {
  cli();                  //< Disable interrupts, so that t1_count_a is not accessed inside an ISR.
  int value = t1_count_a; //< Read t1_count_a value.
  sei();                  //< Enable interrupts.
  return value;           //< Return the read value.
}

/**
 * This is an atomic READ operation and it is intended to be used outside ISRs,
 * for example, in the loop() function.
 */
int64_t read_t1_count_b() {
  cli();                  //< Disable interrupts, so that t1_count_b is not accessed inside an ISR.
  int value = t1_count_b; //< Read t1_count_b value.
  sei();                  //< Enable interrupts.
  return value;           //< Return the read value.
}

/**
 * This is an atomic WRITE operation and it is intended to be used outside ISRs,
 * for example, in the loop() function.
 *
 * @note this functions is not used in this demo program, but shows how to do an atomic write.
 */
void write_t1_count_a(int64_t value) {
  cli();                //< Disable interrupts, so that t1_count is not accessed inside an ISR.
  t1_count_a = value;   //< Update t1_count_a with value.
  sei();                //< Enable interrupts.
}

/**
 * This is an atomic WRITE operation and it is intended to be used outside ISRs,
 * for example, in the loop() function.
 *
 * @note this functions is not used in this demo program, but shows how to do an atomic write.
 */
void write_t1_count_b(int64_t value) {
  cli();                //< Disable interrupts, so that t1_count is not accessed inside an ISR.
  t1_count_b = value;   //< Update t1_count_b with value.
  sei();                //< Enable interrupts.
}

/**
 * F_top := Frequency of IRQs = Frequency of counter to reach TOP
 *
 *             16 000 000 Hz
 * F_top  = -------------------
 *              N * (TOP+1)
 *
 * P_top := Period of counter to reach TOP
 *
 * P_top  = 1 / F_out
 */
static void timer_setup_ctc() {
  // Disable interrupts clearing the I-bit in the Status Register
  cli();

  // 1. Pin Configuration
  // Configure the D9 (OC1A) or D10 (OC1B) or D11 (OC1C) as OUTput pin via DDR.
  // I will use D9 & D10 in this example.
  sbi(D9_DDRX,  D9_DDRXn);
  sbi(D10_DDRX, D10_DDRXn);
  // I want to toggle both D9 & D10 pin status at compare match:
  // Channel A (D9):
  cbi(TCCR1A, COM1A1);
  sbi(TCCR1A, COM1A0);
  // Channel B (D10):
  cbi(TCCR1A, COM1B1);
  sbi(TCCR1A, COM1B0);

  // 2. Clock Configuration
  // A clock (other than CS1[2:0]=000) must be selected to activate the Timer1
  // In this example, we use:
  // - In clk:        16 MHz
  // - Prescaler (N): 1024 (CS1[2:0]=0b101)
  // - Out clk:       15.625 KHz
  // - TOP:           1561
  sbi(TCCR1B, CS12);
  cbi(TCCR1B, CS11);
  sbi(TCCR1B, CS10);

  // 3. Mode Configuaration (Table 14-5)
  // WGM1[3:0] = 0100. Configuration in CTC mode where TOP = OCRnA.
  cbi(TCCR1B, WGM13);
  sbi(TCCR1B, WGM12);
  cbi(TCCR1A, WGM11);
  cbi(TCCR1A, WGM10);
  // TOP value is OCR1A:
  OCR1A = 1561;

  // 4. IRQ Configuration
  // The corresponding interrupt is executed if a compare match in Timer/Counter1 occurs,
  // i.e., when the OCF1A bit is set in the Timer/Counter 1 Interrupt Flag Register – TIFR1.
  // TIMSK1{OCIE1A} must be set to enable IRQ at compare match (TCNT1 == OCR1A).
  sbi(TIMSK1, OCIE1A); // Channel A
  sbi(TIMSK1, OCIE1B); // Channel B

  // Enable interrupts setting the I-bit in the Status Register
  sei();
}

/**
 * @brief The ISR is invoked in when TCNT1 == OCR1A.
 */
ISR(TIMER1_COMPA_vect) {
  t1_count_a++;
  led_toggle();
}

/**
 * @brief The ISR is invoked in when TCNT1 == OCR1B.
 */
ISR(TIMER1_COMPB_vect) {
  t1_count_b++;
}

/***************************************************************************//**
 * A R D U I N O   F U N C T I O N S
 ******************************************************************************/

/** @brief Runs once when you press reset or power the board
 */
void setup() {
  Serial.begin(57600);  //< Configure the serial port
  led_setup();          //< Configure the built-in LED
  delay(5000);          //< So that Serial Monitor window has time to initialize
  timer_setup_ctc();    //< Configure the Timer1
}

/** @brief Runs over and over again forever
 */
void loop() {
  char msg[81];
  int64_t count = read_t1_count_a();
  snprintf(msg, 81, "[Loop: CTC] My counter_a value is %"PRIu16"\n", t1_count_a);
  Serial.print(msg);
  count = read_t1_count_b();
  snprintf(msg, 81, "[Loop: CTC] My counter_b value is %"PRIu16"\n", t1_count_b);
  Serial.print(msg);
}